import json
from collections import defaultdict
from functools import cached_property
from pathlib import Path
from typing import Union, List

import aiofiles
from loguru import logger

from wizwalker import utils
from .client import Client
from .wad import Wad


class WizWalker:
    """
    Represents the main program and handles all windows
    """

    def __init__(self):
        self.window_handles = []
        self.clients = []

        self.wad_cache = None
        self.template_ids = None
        self.node_cache = None

    def __repr__(self):
        return f"<WizWalker {self.window_handles=} {self.clients=}>"

    @cached_property
    def cache_dir(self):
        """
        The dir parsed data is stored in
        """
        return utils.get_cache_folder()

    @cached_property
    def install_location(self) -> Path:
        """
        Wizard101 install location
        """
        return utils.get_wiz_install()

    async def get_template_ids(self) -> dict:
        """
        Loads template ids from cache

        Raises:
            RuntimeError: template ids haven't been cached yet

        Returns:
            the loaded template ids
        """
        try:
            async with aiofiles.open(self.cache_dir / "template_ids.json") as fp:
                message_data = await fp.read()
        except OSError:
            raise RuntimeError(
                "template_ids not yet cached, please run .run or .cache_data"
            )
        else:
            return json.loads(message_data)

    async def get_wad_cache(self) -> dict:
        """
        Gets the wad cache data

        Returns:
            a dict with the current cache data
        """
        try:
            async with aiofiles.open(self.cache_dir / "wad_cache.data") as fp:
                data = await fp.read()
        except OSError:
            data = None

        wad_cache = defaultdict(lambda: defaultdict(lambda: -1))

        if data:
            wad_cache_data = json.loads(data)

            # this is so the default dict inside the default dict isn't replaced
            # by .update
            for k, v in wad_cache_data.items():
                for k1, v1 in v.items():
                    wad_cache[k][k1] = v1

        return wad_cache

    async def write_wad_cache(self):
        """
        Writes wad cache to disk
        """
        async with aiofiles.open(self.cache_dir / "wad_cache.data", "w+") as fp:
            json_data = json.dumps(self.wad_cache)
            await fp.write(json_data)

    async def get_node_cache(self) -> dict:
        """
        Loads the node cache from disk

        Returns:
            The current node cache data
        """
        try:
            async with aiofiles.open(self.cache_dir / "node_cache.data") as fp:
                data = await fp.read()
        except OSError:
            data = None

        node_cache = defaultdict(lambda: 0)

        if data:
            node_cache_data = json.loads(data)
            node_cache.update(node_cache_data)

        return node_cache

    async def write_node_cache(self):
        """
        Writes node cache to disk
        """
        async with aiofiles.open(self.cache_dir / "node_cache.data", "w+") as fp:
            json_data = json.dumps(self.node_cache)
            await fp.write(json_data)

    def get_clients(self):
        """
        Gets all clients currently running

        These are stored in self.clients
        """
        self.get_handles()
        self.clients = [Client(handle) for handle in self.window_handles]

    async def close(self):
        """
        Closes the application and all clients
        """
        for client in self.clients:
            await client.close()

    async def cache_data(self):
        """
        Caches various file data
        """
        root_wad = Wad.from_game_data("Root")

        logger.debug("Caching template")
        await self._cache_template(root_wad)
        # logger.debug("Caching nodes")
        # await self._cache_nodes()

        await self.write_wad_cache()

    async def _cache_template(self, root_wad):
        template_file = "TemplateManifest.xml"

        template_file = await self.check_updated(root_wad, template_file)

        if template_file:
            file_data = await root_wad.get_file("TemplateManifest.xml")
            pharsed_template_ids = utils.pharse_template_id_file(file_data)
            del file_data

            async with aiofiles.open(self.cache_dir / "template_ids.json", "w+") as fp:
                json_data = json.dumps(pharsed_template_ids)
                await fp.write(json_data)

    async def _cache_nodes(self):
        self.node_cache = await self.get_node_cache()

        game_data = self.install_location / "Data" / "GameData"
        all_wads = game_data.glob("*.wad")

        new_node_data = {}
        for wad_name in all_wads:
            wad_name = wad_name.name
            if not self.node_cache[wad_name]:
                logger.debug(f"Checking {wad_name} for node data")
                wad = Wad.from_game_data(wad_name)
                self.node_cache[wad_name] = 1

                try:
                    file_info = await wad.get_file_info("pathNodeData.bin")
                    if file_info.size == 20:
                        continue

                    node_data = await wad.get_file("pathNodeData.bin")
                except RuntimeError:
                    continue
                else:
                    pharsed_node_data = utils.pharse_node_data(node_data)
                    new_node_data[wad_name] = pharsed_node_data

        if new_node_data:
            node_data_json = self.cache_dir / "node_data.json"
            if node_data_json.exists():
                async with aiofiles.open(node_data_json) as fp:
                    json_data = await fp.read()
                    old_node_data = json.loads(json_data)

            else:
                old_node_data = {}

            old_node_data.update(new_node_data)

            async with aiofiles.open(node_data_json, "w+") as fp:
                json_data = json.dumps(old_node_data)
                await fp.write(json_data)

        await self.write_node_cache()

    async def check_updated(self, wad_file: Wad, files: Union[List[str], str]):
        """
        Checks if some wad files have changed since we last accessed them
        """
        if isinstance(files, str):
            files = [files]

        if not self.wad_cache:
            self.wad_cache = await self.get_wad_cache()
            logger.debug(f"TEMP {self.wad_cache=} {type(self.wad_cache)}")

        res = []

        for file_name in files:
            file_info = await wad_file.get_file_info(file_name)

            if self.wad_cache[wad_file.name][file_name] != file_info.size:
                logger.debug(
                    f"{file_name} has updated. old: {self.wad_cache[wad_file.name][file_name]} new: {file_info.size}"
                )
                res.append(file_name)
                self.wad_cache[wad_file.name][file_name] = file_info.size

            else:
                logger.debug(f"{file_name} has not updated from {file_info.size}")

        return res

    def run(self):
        self.get_clients()
        self.cache_data()

    @staticmethod
    def start_wiz_client():
        utils.start_instance()

    def get_handles(self):
        """
        Gets all the wizard handles

        get_clients should be called over this
        """
        current_handles = utils.get_all_wizard_handles()

        if not current_handles:
            raise RuntimeError("No handles found")

        self.window_handles = current_handles



import ctypes
from enum import Enum


user32 = ctypes.windll.user32


# noinspection PyPep8
class Keycode(Enum):
    Left_mouse = 1
    Right_mouse = 2
    Control_break_processing = 3
    Middle_mouse = 4
    X1_mouse = 5
    X2_mouse = 6
    Undefined = 7
    BACKSPACE = 8
    TAB = 9
    CLEAR = 12
    ENTER = 13
    SHIFT = 16
    CTRL = 17
    ALT = 18
    PAUSE = 19
    CAPS_LOCK = 20
    ESC = 27
    SPACEBAR = 32
    PAGE_UP = 33
    PAGE_DOWN = 34
    END = 35
    HOME = 36
    LEFT_ARROW = 37
    UP_ARROW = 38
    RIGHT_ARROW = 39
    DOWN_ARROW = 40
    SELECT = 41
    PRINT = 42
    EXECUTE = 43
    PRINT_SCREEN = 44
    INS = 45
    DEL = 46
    HELP = 47
    ZERO = 48
    ONE = 49
    TWO = 50
    THREE = 51
    FOUR = 52
    FIVE = 53
    SIX = 54
    SEVEN = 55
    EIGHT = 56
    NINE = 57
    A = 65
    B = 66
    C = 67
    D = 68
    E = 69
    F = 70
    G = 71
    H = 72
    I = 73
    J = 74
    K = 75
    L = 76
    M = 77
    N = 78
    O = 79
    P = 80
    Q = 81
    R = 82
    S = 83
    T = 84
    U = 85
    V = 86
    W = 87
    X = 88
    Y = 89
    Z = 90
    Left_Windows = 91
    Right_Windows = 92
    Applications = 93
    Reserved = 252
    Computer_Sleep = 95
    Numeric_pad_0 = 96
    Numeric_pad_1 = 97
    Numeric_pad_2 = 98
    Numeric_pad_3 = 99
    Numeric_pad_4 = 100
    Numeric_pad_5 = 101
    Numeric_pad_6 = 102
    Numeric_pad_7 = 103
    Numeric_pad_8 = 104
    Numeric_pad_9 = 105
    Multiply = 106
    Add = 107
    Separator = 108
    Subtract = 109
    Decimal = 110
    Divide = 111
    F1 = 112
    F2 = 113
    F3 = 114
    F4 = 115
    F5 = 116
    F6 = 117
    F7 = 118
    F8 = 119
    F9 = 120
    F10 = 121
    F11 = 122
    F12 = 123
    F13 = 124
    F14 = 125
    F15 = 126
    F16 = 127
    F17 = 128
    F18 = 129
    F19 = 130
    F20 = 131
    F21 = 132
    F22 = 133
    F23 = 134
    F24 = 135
    NUM_LOCK = 144
    SCROLL_LOCK = 145
    Left_SHIFT = 160
    Right_SHIFT = 161
    Left_CONTROL = 162
    Right_CONTROL = 163
    Left_MENU = 164
    Right_MENU = 165
    Browser_Back = 166
    Browser_Forward = 167
    Browser_Refresh = 168
    Browser_Stop = 169
    Browser_Search = 170
    Browser_Favorites = 171
    Browser_Start_and_Home = 172
    Volume_Mute = 173
    Volume_Down = 174
    Volume_Up = 175
    Next_Track = 176
    Previous_Track = 177
    Stop_Media = 178
    Play_Pause_Media = 179
    Start_Mail = 180
    Select_Media = 181
    Start_Application_1 = 182
    Start_Application_2 = 183
    OEM_specific = 230
    IME_PROCESS = 229
    Attn = 246
    CrSel = 247
    ExSel = 248
    Erase_EOF = 249
    Play = 250
    Zoom = 251
    PA1 = 253
    Clear = 254



import logging

from loguru import logger

# TODO: 1.0 remove windows folder and expose memory
from .constants import *
from .wad import Wad
from .nif import NifMap
from .errors import *
from . import utils, windows, cli
from .application import WizWalker
from .client import Client

logger.disable("wizwalker")
logging.getLogger("pymem").setLevel(logging.FATAL)



from typing import Union


class WizWalkerError(Exception):
    """
    Base wizwalker exception, all exceptions raised should inharit from this
    """

    pass


class HookPatternFailed(WizWalkerError):
    """
    Raised when the pattern scan for a hook fails
    """

    def __init__(self):
        super().__init__(
            "A hook search pattern failed. You most likely need to restart the client"
        )


class HookNotActive(WizWalkerError):
    """
    Raised when doing something that requires a hook to be active
    but it is not

    Attributes:
        hook: The hook that is not active
    """

    def __init__(self, hook: str):
        super().__init__(f"{hook} is not active.")
        self.hook = hook


class HookAlreadyActivated(WizWalkerError):
    """
    Raised when trying to activate an active hook

    Attributes:
        hook: The hook that is already active
    """

    def __init__(self, hook: str):
        super().__init__(f"{hook} was already activated.")
        self.hook = hook


class NotEnoughPips(WizWalkerError):
    """
    Raised when trying to use a card that costs more pips then
    are available

    Attributes:
        missing: The amount of missing pips
    """

    def __init__(self, missing: int):
        super().__init__(f"Missing {missing} pips needed.")
        self.missing = missing


class NotEnoughMana(WizWalkerError):
    """
    Raised when trying to use a card that cost more mana than
    is available

    Attributes:
        missing: The amount of missing mana
    """

    def __init__(self, missing: int):
        super().__init__(f"Missing {missing} mana needed.")
        self.missing = missing


class CardNotFound(WizWalkerError):
    """
    Raised when searching for a card brings no results

    Attributes:
        card_name_or_position: The unfound card name or position
    """

    def __init__(self, card_name_or_positon: Union[str, int]):
        super().__init__(
            f"Card with name or position {card_name_or_positon} not found."
        )
        self.card_name_or_positon = card_name_or_positon


class CardAlreadyEnchanted(WizWalkerError):
    """
    Raised when trying to encahnt an already enchanted card
    """

    def __init__(self):
        super().__init__("That card is already enchanted.")


class HotkeyAlreadyRegistered(WizWalkerError):
    def __init__(self, key: str):
        super().__init__(f"{key} already registered")



import asyncio
import ctypes
import functools
import io
import math
import struct
import subprocess
import winreg
import zlib
from collections import namedtuple
from concurrent.futures.thread import ThreadPoolExecutor
from ctypes import WinDLL
from pathlib import Path
from typing import Callable, Iterable
from xml.etree import ElementTree

import appdirs

from wizwalker.constants import Keycode

user32 = WinDLL("user32")


XYZ = namedtuple("XYZ", "x, y, z")


def _xyz_sub(self, other):
    if not isinstance(other, XYZ):
        raise ValueError(f"XYZ can only be subtracted from XYZ not {type(other)}")

    return math.dist((self.x, self.y), (other.x, other.y))


XYZ.__sub__ = _xyz_sub


# Modified from https://github.com/Gorialis/jishaku/blob/master/jishaku/functools.py#L19
# This license covers the below function
# MIT License
#
# Copyright (c) 2020 Devon (Gorialis) R
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
def executor_function(sync_function: Callable):
    @functools.wraps(sync_function)
    async def sync_wrapper(*args, **kwargs):
        loop = asyncio.get_event_loop()
        internal_function = functools.partial(sync_function, *args, **kwargs)

        with ThreadPoolExecutor() as pool:
            return await loop.run_in_executor(pool, internal_function)

    return sync_wrapper


def get_wiz_install() -> Path:
    """
    Get the game install root dir
    """
    reg = winreg.ConnectRegistry(None, winreg.HKEY_CURRENT_USER)

    try:
        key = winreg.OpenKey(
            reg,
            r"Software\Microsoft\Windows\CurrentVersion\Uninstall\{A9E27FF5-6294-46A8-B8FD-77B1DECA3021}",
            0,
            winreg.KEY_READ,
        )
    except OSError:
        raise Exception("Wizard101 install not found, do you have it installed?")

    install_location = Path(winreg.QueryValueEx(key, "InstallLocation")[0]).absolute()
    return install_location


def start_instance():
    """
    Starts a wizard101 instance
    """
    location = get_wiz_install()
    subprocess.Popen(
        rf"{location}\Bin\WizardGraphicalClient.exe -L login.us.wizard101.com 12000",
        cwd=rf"{location}\Bin",
    )


def instance_login(window_handle: int, username: str, password: str):
    """
    Login to an instance on the login screen

    Args:
        window_handle: Handle to window
        username: Username to login with
        password: Password to login with
    """

    def send_chars(chars: str):
        for char in chars:
            user32.SendMessageW(window_handle, 0x102, ord(char), 0)

    send_chars(username)
    # tab
    user32.SendMessageW(window_handle, 0x102, 9, 0)
    send_chars(password)
    # enter
    user32.SendMessageW(window_handle, 0x102, 13, 0)


async def start_instances_with_login(instance_number: int, logins: Iterable):
    """
    Start a number of instances and login to them with logins

    Args:
        instance_number: number of instances to start
        logins: logins to use
    """
    start_handles = set(get_all_wizard_handles())

    for _ in range(instance_number):
        start_instance()

    # TODO: have way to properly check if instances are on login screen
    # waiting for instances to start
    await asyncio.sleep(7)

    new_handles = set(get_all_wizard_handles()).difference(start_handles)

    for handle, username_password in zip(new_handles, logins):
        username, password = username_password.split(":")
        instance_login(handle, username, password)


def calculate_perfect_yaw(current_xyz: XYZ, target_xyz: XYZ) -> float:
    """
    Calculates the perfect yaw to reach an xyz in a stright line

    Args:
        current_xyz: Starting position xyz
        target_xyz: Ending position xyz
    """
    target_line = math.dist(
        (current_xyz.x, current_xyz.y), (target_xyz.x, target_xyz.y)
    )
    origin_line = math.dist(
        (current_xyz.x, current_xyz.y), (current_xyz.x, current_xyz.y - 1)
    )
    target_to_origin_line = math.dist(
        (target_xyz.x, target_xyz.y), (current_xyz.x, current_xyz.y - 1)
    )
    # target_angle = math.cos(origin_line / target_line)
    target_angle = math.acos(
        (pow(target_line, 2) + pow(origin_line, 2) - pow(target_to_origin_line, 2))
        / (2 * origin_line * target_line)
    )

    if target_xyz.x > current_xyz.x:
        # outside
        target_angle_degres = math.degrees(target_angle)
        perfect_yaw = math.radians(360 - target_angle_degres)
    else:
        # inside
        perfect_yaw = target_angle

    return perfect_yaw


def get_cache_folder() -> Path:
    """
    Get the wizwalker cache folder
    """
    app_name = "WizWalker"
    app_author = "StarrFox"
    cache_dir = Path(appdirs.user_cache_dir(app_name, app_author))

    if not cache_dir.exists():
        cache_dir.mkdir(parents=True)

    return cache_dir


def get_logs_folder() -> Path:
    """
    Get the wizwalker log folder
    """
    app_name = "WizWalker"
    app_author = "StarrFox"
    log_dir = Path(appdirs.user_log_dir(app_name, app_author))

    if not log_dir.exists():
        log_dir.mkdir(parents=True)

    return log_dir


def get_all_wizard_handles() -> list:
    """
    Get handles to all currently open wizard clients
    """
    target_class = "Wizard Graphical Client"

    def callback(handle):
        class_name = ctypes.create_unicode_buffer(len(target_class))
        user32.GetClassNameW(handle, class_name, len(target_class) + 1)
        if target_class == class_name.value:
            return True

    return get_windows_from_predicate(callback)


def get_windows_from_predicate(predicate: Callable) -> list:
    """
    Get all windows that match a predicate

    Args:
        predicate: the predicate windows should pass

    Examples:
        .. code-block:: py

            def predicate(window_handle):
                if window_handle == 0:
                    # handle will be returned
                    return True
                else:
                    # handle will not be returned
                    return False
    """
    handles = []

    def callback(handle, _):
        if predicate(handle):
            handles.append(handle)

        # iterate all windows, (True)
        return 1

    enumwindows_func_type = ctypes.WINFUNCTYPE(
        ctypes.c_bool, ctypes.c_int, ctypes.POINTER(ctypes.c_int),
    )

    callback = enumwindows_func_type(callback)
    user32.EnumWindows(callback, 0)

    return handles


def pharse_template_id_file(file_data: bytes) -> dict:
    """
    Pharse a template id file's data
    """
    if not file_data.startswith(b"BINd"):
        raise RuntimeError("No BINd id string")

    data = zlib.decompress(file_data[0xD:])

    total_size = len(data)
    data = io.BytesIO(data)

    data.seek(0x24)

    out = {}
    while data.tell() < total_size:
        size = ord(data.read(1)) // 2

        string = data.read(size).decode()
        data.read(8)  # unknown bytes

        # Little endian int
        entry_id = struct.unpack("<i", data.read(4))[0]

        data.read(0x10)  # next entry

        out[entry_id] = string

    return out


def pharse_node_data(file_data: bytes) -> dict:
    """
    Converts data into a dict of node nums to points
    """
    entry_start = b"\xFE\xDB\xAE\x04"

    node_data = {}
    # no nodes
    if len(file_data) == 20:
        return node_data

    # header
    file_data = file_data[20:]

    last_start = 0
    while file_data:
        start = file_data.find(entry_start, last_start)
        if start == -1:
            break

        entry = file_data[start : start + 48 + 2]

        cords_data = entry[16 : 16 + (4 * 3)]
        x = struct.unpack("<f", cords_data[0:4])[0]
        y = struct.unpack("<f", cords_data[4:8])[0]
        z = struct.unpack("<f", cords_data[8:12])[0]

        node_num = entry[48 : 48 + 2]
        unpacked_num = struct.unpack("<H", node_num)[0]

        node_data[unpacked_num] = (x, y, z)

    return node_data


async def timed_send_key(window_handle: int, key: Keycode, seconds: float):
    """
    Send a key for a number of seconds

    Args:
        window_handle: Handle to window to send key to
        key: The key to send
        seconds: Number of seconds to send the key
    """
    keydown_task = asyncio.create_task(_send_keydown_forever(window_handle, key))
    await asyncio.sleep(seconds)
    keydown_task.cancel()
    user32.SendMessageW(window_handle, 0x101, key.value, 0)


async def _send_keydown_forever(window_handle: int, key: Keycode):
    while True:
        user32.SendMessageW(window_handle, 0x100, key.value, 0)
        await asyncio.sleep(0.05)



import asyncio
import re
import threading
import sys
from typing import Any, Coroutine, Union


import terminaltables
import aioconsole
from aiomonitor import Monitor, start_monitor, cli
from aiomonitor.utils import close_server, console_proxy


def init_console_server(host: str, port: int, _locals, loop):
    def _factory(streams=None) -> aioconsole.AsynchronousConsole:
        return NoBannerConsole(locals=_locals, streams=streams, loop=loop)

    coro = aioconsole.start_interactive_server(
        host=host, port=port, factory=_factory, loop=loop
    )
    console_future = asyncio.run_coroutine_threadsafe(coro, loop=loop)
    return console_future


class NoBannerConsole(aioconsole.AsynchronousConsole):
    async def _interact(self, banner=None):
        # Get ps1 and ps2
        try:
            sys.ps1
        except AttributeError:
            sys.ps1 = ">>> "
        try:
            sys.ps2
        except AttributeError:
            sys.ps2 = "... "
        # Run loop
        more = 0
        while 1:
            try:
                if more:
                    prompt = sys.ps2
                else:
                    prompt = sys.ps1
                try:
                    line = await self.raw_input(prompt)
                except EOFError:
                    self.write("\n")
                    await self.flush()
                    break
                else:
                    more = await self.push(line)
            except asyncio.CancelledError:
                self.write("\nKeyboardInterrupt\n")
                await self.flush()
                self.resetbuffer()
                more = 0


class WizWalkerConsole(Monitor):
    intro = "WizWalkerCLI\n{tasknum} task{s} running. Use help (?) for commands.\n"
    prompt = "WW > "

    def write(self, message: str):
        self._sout.write(message + "\n")
        self._sout.flush()

    def get_local(self, name: str) -> Any:
        res = self._locals.get(name)
        if res is None:
            raise ValueError(f"{name} not in locals")

        return res

    def run_coro(self, coro: Coroutine, timeout: Union[int, None] = 10):
        future = asyncio.run_coroutine_threadsafe(coro, self._loop)

        try:
            result = future.result(timeout)
        except asyncio.TimeoutError:
            self.write(f"Timeout error with coro {coro.__name__}")
            future.cancel()
        except Exception as exc:
            import traceback

            traceback.print_exception(type(exc), exc, exc.__traceback__)
        else:
            return result

    def do_console(self):
        """Switch to async Python REPL"""
        if not self._console_enabled:
            self.write("Python console disabled for this session")
            return

        fut = init_console_server(
            self._host, self._console_port, self._locals, self._loop
        )
        server = fut.result(timeout=3)
        try:
            console_proxy(self._sin, self._sout, self._host, self._console_port)
        finally:
            coro = close_server(server)
            close_fut = asyncio.run_coroutine_threadsafe(coro, loop=self._loop)
            close_fut.result(timeout=15)

    def do_start(self):
        """Attach and hook to all clients"""
        walker = self.get_local("walker")
        walker.get_clients()
        self.write(f"Attached to {len(walker.clients)} clients")
        for idx, client in enumerate(walker.clients):
            self.run_coro(client.activate_hooks())
            self.write(f"client-{idx}: hooked all")

    def do_exit(self) -> None:
        """exit walker and re-write hooks"""
        walker = self.get_local("walker")
        self.write("Closing wizwalker, hooks should be rewritten")
        self.run_coro(walker.close())

    def do_info(self):
        """Print out info from each client"""
        walker = self.get_local("walker")

        for idx, client in enumerate(walker.clients):
            table_data = [["attribute", "value"]]

            # TODO: add player_base and player_stat_base and all other base address (optional arg)
            client_attrs = [
                "xyz",
                "yaw",
                "roll",
                "pitch",
                "scale",
                "quest_xyz",
                "health",
                "mana",
                "potions",
                "gold",
                "energy",
                "backpack_space_used",
                "backpack_space_total",
                "move_lock",
            ]
            for attr in client_attrs:
                table_data.append([attr, self.run_coro(getattr(client, attr)())])

            table = terminaltables.AsciiTable(table_data, f"client-{idx}")
            self.write(table.table)

    def do_cache(self):
        """Cache data"""
        walker = self.get_local("walker")
        self.run_coro(walker.cache_data())
        self.write("Cached data")

    def do_teleport(self, x: float, y: float, z: float = None, yaw: float = None):
        """Teleport to a location

        x y z and yaw are all optional
        """
        walker = self.get_local("walker")
        for client in walker.clients:
            self.run_coro(client.teleport(x=x, y=y, z=z, yaw=yaw))

        self.write("Teleported")

    def do_goto(self, x: float, y: float):
        """Go to a location in the world"""
        walker = self.get_local("walker")
        for client in walker.clients:
            self.run_coro(client.goto(x, y))

        self.write("Completed goto")

    def do_getid(self, pattern: str):
        """Get templateid items that match a pattern

        cache command must be run first
        """
        walker = self.get_local("walker")
        template_ids = self.run_coro(walker.get_template_ids())

        regex = re.compile(pattern, re.IGNORECASE)
        for tid, name in template_ids.items():
            if regex.match(name):
                self.write(f"{tid=} {name=}")

    def do_checkid(self, tid: str):
        """Get the name mapped to a templateid

        cache command must be run first
        """
        walker = self.get_local("walker")
        template_ids: dict = self.run_coro(walker.get_template_ids())

        try:
            self.write(f"{tid} => {template_ids[tid]}")
        except KeyError:
            self.write(f"No item with id {tid}")

    def do_click(self, x: int, y: int):
        """Click a certain x, y"""
        walker = self.get_local("walker")
        for client in walker.clients:
            self.run_coro(client.click(x, y))

        self.write("Completed click")

    def do_onehook(self, name: str):
        """Get clients and activate one hook"""
        walker = self.get_local("walker")
        walker.get_clients()
        self.write(f"Attached to {len(walker.clients)} clients")

        for client in walker.clients:
            self.run_coro(client.activate_hooks(name))

        self.write(f"Hooked {name} on all clients")


def test_monitor():
    cli.monitor_client(cli.MONITOR_HOST, cli.MONITOR_PORT)


# TODO: replace app with walker when WizWalker has run loop
def start_console(loop=None, locals=None):
    if loop is None:
        loop = asyncio.get_event_loop()

    def app():
        try:
            loop.run_forever()
        except KeyboardInterrupt:
            print("Closing wizwalker; hooks should be rewritten")
            loop.run_until_complete(locals["walker"].close())

            tasks = asyncio.Task.all_tasks(loop)
            for task in tasks:
                if not task.done():
                    task.cancel()

            loop.run_until_complete(asyncio.gather(*tasks, return_exceptions=True))

    def run_monitor():
        cli.monitor_client(cli.MONITOR_HOST, cli.MONITOR_PORT)

    monitor_thread = threading.Thread(target=run_monitor, daemon=True)
    monitor_thread.start()

    # monitor_proc = multiprocessing.Process(target=test_monitor, daemon=True)
    # loop.call_later(2, monitor_proc.start)

    with start_monitor(loop, monitor=WizWalkerConsole, locals=locals):
        app()



import re
import struct
import ctypes.wintypes
from typing import Any, Tuple, Callable, Type, List

import pymem
import pymem.pattern
import pymem.ressources.kernel32

from wizwalker import HookPatternFailed


def pack_to_int_or_longlong(num: int) -> bytes:
    try:
        return struct.pack("<i", num)
    except struct.error:
        return struct.pack("<q", num)


class MEMORY_BASIC_INFORMATION(ctypes.Structure):
    _fields_ = [
        ("BaseAddress", ctypes.c_void_p),
        ("AllocationBase", ctypes.c_void_p),
        ("AllocationProtect", ctypes.wintypes.DWORD),
        ("PartitionId", ctypes.wintypes.WORD),
        ("RegionSize", ctypes.c_size_t),
        ("State", ctypes.wintypes.DWORD),
        ("Protect", ctypes.wintypes.DWORD),
        ("Type", ctypes.wintypes.DWORD),
    ]


def scan_all_from(start_address: int, handle: int, pattern: bytes):
    next_region = start_address
    found = None

    while next_region < 0x7FFFFFFF0000:
        next_region, found = pymem.pattern.scan_pattern_page(
            handle, next_region, pattern
        )
        if found:
            break

    return found


class MemoryHook:
    def __init__(self, memory_handler):
        self.memory_handler = memory_handler
        self.jump_original_bytecode = None

        self.hook_address = None
        self.jump_address = None

        self.jump_bytecode = None
        self.hook_bytecode = None

        # so we can dealloc it on unhook
        self._allocated_addresses = []

    def alloc(self, size: int) -> int:
        """
        Allocate <size> bytes
        """
        addr = self.memory_handler.process.allocate(size)
        self._allocated_addresses.append(addr)
        return addr

    def prehook(self):
        """
        Called after bytecode is prepared and before written
        """
        pass

    def posthook(self):
        """
        Called after bytecode is written
        """
        pass

    def pattern_scan(self, pattern: bytes, *, module: str = None):
        if module:
            module = pymem.process.module_from_name(
                self.memory_handler.process.process_handle, module
            )
            jump_address = pymem.pattern.pattern_scan_module(
                self.memory_handler.process.process_handle, module, pattern
            )

        else:
            jump_address = scan_all_from(
                self.memory_handler.process.process_base.lpBaseOfDll,
                self.memory_handler.process.process_handle,
                pattern,
            )

        # TODO: maybe error if None?

        return jump_address

    def read_bytes(self, address: int, size: int) -> bytes:
        return self.memory_handler.process.read_bytes(address, size)

    def write_bytes(self, address: int, _bytes: bytes):
        self.memory_handler.process.write_bytes(
            address, _bytes, len(_bytes),
        )

    def get_jump_address(self, pattern: bytes, module: str = None) -> int:
        """
        gets the address to write jump at
        """
        jump_address = self.pattern_scan(pattern, module=module)

        if jump_address is None:
            raise HookPatternFailed()

        return jump_address

    def get_hook_address(self, size: int) -> int:
        return self.alloc(size)

    def get_jump_bytecode(self) -> bytes:
        """
        Gets the bytecode to write to the jump address
        """
        raise NotImplemented()

    def get_hook_bytecode(self) -> bytes:
        """
        Gets the bytecord to write to the hook address
        """
        raise NotImplemented()

    def get_pattern(self) -> Tuple[bytes, str]:
        raise NotImplemented()

    def hook(self) -> Any:
        """
        Writes jump_bytecode to jump address and hook bytecode to hook address
        """
        pattern, module = self.get_pattern()

        self.jump_address = self.get_jump_address(pattern, module=module)
        self.hook_address = self.get_hook_address(50)

        self.hook_bytecode = self.get_hook_bytecode()
        self.jump_bytecode = self.get_jump_bytecode()

        self.jump_original_bytecode = self.read_bytes(
            self.jump_address, len(self.jump_bytecode)
        )

        self.prehook()

        self.write_bytes(self.hook_address, self.hook_bytecode)
        self.write_bytes(self.jump_address, self.jump_bytecode)

        self.posthook()

    def unhook(self):
        """
        Deallocates hook memory and rewrites jump addr to it's origional code,
        also called when a client is closed
        """
        self.write_bytes(self.jump_address, self.jump_original_bytecode)
        # TODO: replace for function usage hook
        for addr in self._allocated_addresses:
            self.memory_handler.process.free(addr)


class AutoBotBaseHook(MemoryHook):
    """
    Subclass of MemoryHook that uses an autobot function for bytes so addresses arent huge
    """

    # Yes this does have to be this long
    AUTOBOT_PATTERN = (
        rb"\x48\x8B\xC4\x55\x41\x54\x41\x55\x41\x56\x41\x57"
        rb"\x48\x8D\xA8\x18\xFE\xFF\xFF\x48\x81\xEC\xC0\x02\x00"
        rb"\x00\x48\xC7\x45\xC8\xFE\xFF\xFF\xFF\x48\x89\x58\x10"
        rb"\x48\x89\x70\x18\x48\x89\x78\x20\x48\x8B\x05\x63\xA0"
        rb"\xBE\x01\x48\x33\xC4\x48\x89\x85\xB0\x01\x00\x00\x4C\x8B\xE9"
    )
    # rounded down
    AUTOBOT_SIZE = 3900

    _autobot_addr = None
    # How far into the function we are
    _autobot_bytes_offset = 0

    _autobot_original_bytes = None

    # this is really hacky
    _hooked_instances = 0

    def alloc(self, size: int) -> int:
        if self._autobot_addr is None:
            addr = self.pattern_scan(self.AUTOBOT_PATTERN)
            # this is so all instances have the address
            AutoBotBaseHook._autobot_addr = addr

        if self._autobot_bytes_offset + size > self.AUTOBOT_SIZE:
            raise RuntimeError("Somehow used the entirety of the autobot function")

        if self._autobot_original_bytes is None:
            AutoBotBaseHook._autobot_original_bytes = self.read_bytes(
                self._autobot_addr, self.AUTOBOT_SIZE
            )
            # this is so instructions don't collide
            self.write_bytes(self._autobot_addr, b"\x00" * self.AUTOBOT_SIZE)

        addr = self._autobot_addr + self._autobot_bytes_offset
        AutoBotBaseHook._autobot_bytes_offset += size

        return addr

    def hook(self) -> Any:
        AutoBotBaseHook._hooked_instances += 1
        return super().hook()

    # TODO: make it so this "deallocates" autobot bytes
    # This if overwritten bc we never call free
    def unhook(self):
        AutoBotBaseHook._hooked_instances -= 1
        self.write_bytes(self.jump_address, self.jump_original_bytecode)

        if self._hooked_instances == 0:
            self.write_bytes(self._autobot_addr, self._autobot_original_bytes)
            AutoBotBaseHook._autobot_bytes_offset = 0


# This is a function and not a subclass so I don't have to change anything in handler
def simple_hook(
    *,
    pattern: bytes,
    module: str = "WizardGraphicalClient.exe",
    bytecode_generator: Callable,
    instruction_length: int = 5,
    exports: List[Tuple],
    noops: int = 0,
) -> Type[MemoryHook]:
    """
    Create a simple hook from base args

    Args:
        pattern: The pattern to scan for
        module: The module to scan within (None) for all; defaults to exe
        bytecode_generator: A function that returns the bytecode to write
        instruction_length: length of instructions at jump address
        exports: list of tuples in the form (name, size)
        noops: number of noops to put at jump
    """

    class _memory_hook(AutoBotBaseHook):
        def __init__(self, memory_handler):
            super().__init__(memory_handler)

        def get_pattern(self):
            return pattern, module

        def get_jump_bytecode(self) -> bytes:
            distance = self.hook_address - self.jump_address

            relitive_jump = distance - 5
            packed_relitive_jump = struct.pack("<i", relitive_jump)

            return b"\xE9" + packed_relitive_jump + (b"\x90" * noops)

        def get_hook_bytecode(self) -> bytes:
            packed_exports = []
            for export in exports:
                # addr = self.alloc(export[1])
                addr = self.memory_handler.process.allocate(export[1])
                setattr(self, export[0], addr)
                packed_addr = pack_to_int_or_longlong(addr)
                packed_exports.append((export[0], packed_addr))

            bytecode = bytecode_generator(self, packed_exports)

            return_addr = self.jump_address + instruction_length

            relitive_return_jump = return_addr - (self.hook_address + len(bytecode)) - 5
            packed_relitive_return_jump = struct.pack("<i", relitive_return_jump)

            bytecode += b"\xE9" + packed_relitive_return_jump

            return bytecode

        def unhook(self):
            super().unhook()
            for export in exports:
                if getattr(self, export[0], None):
                    self.memory_handler.process.free(getattr(self, export[0]))

    return _memory_hook


def player_hook_bytecode_gen(_, packed_exports):
    # We use ecx bc we want 4 bytes only
    bytecode = (
        b"\x51"  # push rcx
        b"\x8B\x88\x74\x04\x00\x00"  # mov ecx,[rax+474]
        # check if player
        b"\x83\xF9\x08"  # cmp ecx,08
        b"\x59"  # pop rcx
        b"\x0F\x85\x0A\x00\x00\x00"  # jne 10 down
        # mov(abs) [addr], rax
        b"\x48\xA3" + packed_exports[0][1] +
        # original code
        b"\xF2\x0F\x10\x40\x58"  # movsd xxmo,[rax+58]
    )
    return bytecode


PlayerHook = simple_hook(
    pattern=rb"\xF2\x0F\x10\x40\x58\xF2\x0F\x11\x02",
    bytecode_generator=player_hook_bytecode_gen,
    exports=[("player_struct", 8)],
)


def player_stat_hook_bytecode_gen(_, packed_exports):
    # fmt: off
    bytecode = (
        b"\x50"  # push rax
        b"\x48\x89\xC8"  # mov rax, rcx
        b"\x48\xA3" + packed_exports[0][1] +  # mov qword ptr [stat_export], rax
        b"\x58"  # pop rax
        # original code
        b"\x03\x59\x54"  # add ebx, dword ptr [rcx+0x54]
        b"\x0F\x29\x74\x24\x20"  # movaps [rsp+20],xmm6
    )
    # fmt: on
    return bytecode


PlayerStatHook = simple_hook(
    pattern=rb"\x03\x59\x54\x0F\x29\x74\x24\x20\x0F\x57",
    bytecode_generator=player_stat_hook_bytecode_gen,
    instruction_length=8,
    exports=[("stat_addr", 8)],
)


def quest_hook_bytecode_gen(_, packed_exports):
    # fmt: off
    bytecode = (
        b"\x50"  # push rcx
        b"\x49\x8D\x86\xAC\x0C\x00\x00"  # lea rcx,[r14+CAC]
        b"\x48\xA3" + packed_exports[0][1] +  # mov [export],rcx
        b"\x58"  # pop rcx
        b"\xF3\x41\x0F\x10\x86\xAC\x0C\x00\x00"  # original code
    )
    # fmt: on
    return bytecode


QuestHook = simple_hook(
    pattern=rb"\xF3\x41\x0F\x10\x86\xAC\x0C\x00\x00",
    bytecode_generator=quest_hook_bytecode_gen,
    exports=[("cord_struct", 4)],
    noops=4,
)


def backpack_stat_bytecode_gen(_, packed_exports):
    # fmt: off
    bytecode = (
        b"\x50"  # push rax
        b"\x48\x8D\x83\x38\x05\x00\x00"  # lea rax,[rbx+538]
        b"\x48\xA3" + packed_exports[0][1] +  # mov [export],rax
        b"\x58"  # pop rax
        b"\x01\x83\x38\x05\x00\x00"  # original code
    )
    # fmt: on
    return bytecode


BackpackStatHook = simple_hook(
    pattern=rb"\x01\x83\x38\x05\x00\x00\x48\x8B\x8B\x58\x02\x00\x00\x48\x85\xC9",
    bytecode_generator=backpack_stat_bytecode_gen,
    instruction_length=6,
    exports=[("backpack_struct_addr", 8)],
    noops=1,
)


def duel_hook_bytecode_gen(_, exports):
    # fmt: off
    bytecode = (
        b"\x48\x39\xD1"  # cmp rcx,rdx
        b"\x0F\x85\x0F\x00\x00\x00"  # jne 15
        b"\x50"  # push rax
        b"\x49\x8B\x07"  # mov rax,[r15]
        b"\x48\xA3" + exports[0][1] +  # mov [current_duel],rax
        b"\x58"  # pop rax
        b"\x48\x89\x4C\x24\x50"  # original instruction
    )
    # fmt: on

    return bytecode


DuelHook = simple_hook(
    pattern=rb"\x48\x89\x4C\x24\x50\x4C\x89\x74\x24\x48",
    bytecode_generator=duel_hook_bytecode_gen,
    exports=[("current_duel_addr", 8)],
)


class User32GetClassInfoBaseHook(AutoBotBaseHook):
    """
    Subclass of MemoryHook that uses the user32.GetClassInfoExA for bytes so addresses arent huge
    """

    # Yes this does have to be this long
    AUTOBOT_PATTERN = (
        rb"\x48\x89\x5C\x24\x20\x55\x56\x57\x41\x54\x41\x55\x41\x56\x41\x57\x48\x8D"
        rb"\xAC\x24\x60\xFF\xFF\xFF\x48\x81\xEC\xA0\x01\x00\x00\x48\x8B\x05\x9A\xFD\x0A\x00"
    )
    # rounded down
    AUTOBOT_SIZE = 1200

    _autobot_addr = None
    # How far into the function we are
    _autobot_bytes_offset = 0

    _autobot_original_bytes = None

    # this is really hacky
    _hooked_instances = 0

    def alloc(self, size: int) -> int:
        if self._autobot_addr is None:
            addr = self.pattern_scan(self.AUTOBOT_PATTERN)
            # this is so all instances have the address
            User32GetClassInfoBaseHook._autobot_addr = addr

        if self._autobot_bytes_offset + size > self.AUTOBOT_SIZE:
            raise RuntimeError("Somehow used the entirety of the autobot function")

        if self._autobot_original_bytes is None:
            User32GetClassInfoBaseHook._autobot_original_bytes = self.read_bytes(
                self._autobot_addr, self.AUTOBOT_SIZE
            )
            # this is so instructions don't collide
            self.write_bytes(self._autobot_addr, b"\x00" * self.AUTOBOT_SIZE)

        addr = self._autobot_addr + self._autobot_bytes_offset
        User32GetClassInfoBaseHook._autobot_bytes_offset += size

        return addr

    def hook(self) -> Any:
        User32GetClassInfoBaseHook._hooked_instances += 1
        return super().hook()

    def unhook(self):
        User32GetClassInfoBaseHook._hooked_instances -= 1
        self.write_bytes(self.jump_address, self.jump_original_bytecode)

        if self._hooked_instances == 0:
            self.write_bytes(self._autobot_addr, self._autobot_original_bytes)
            User32GetClassInfoBaseHook._autobot_bytes_offset = 0


class MouselessCursorMoveHook(User32GetClassInfoBaseHook):
    def __init__(self, memory_handler):
        super().__init__(memory_handler)
        self.mouse_pos_addr = None

        self.toggle_bool_addrs = ()

    def posthook(self):
        bool_one_address = self.pattern_scan(
            rb"[\x00\x01]\xFF\x50\x18\x66\xC7", module="WizardGraphicalClient.exe"
        )
        bool_two_address = self.pattern_scan(
            rb"[\x00\x01]\x33\xFF\x89\xBD\x80\x01\x00\x00\x8D",
            module="WizardGraphicalClient.exe",
        )

        if bool_one_address is None or bool_two_address is None:
            raise RuntimeError("toogle bool address pattern failed")

        self.toggle_bool_addrs = (bool_one_address, bool_two_address)
        self.memory_handler.process.write_uchar(bool_one_address, 1)
        self.memory_handler.process.write_uchar(bool_two_address, 1)

    def set_mouse_pos_addr(self):
        self.mouse_pos_addr = self.memory_handler.process.allocate(8)

    def free_mouse_pos_addr(self):
        self.memory_handler.process.free(self.mouse_pos_addr)

    def get_jump_bytecode(self) -> bytes:
        # distance = end - start
        distance = self.hook_address - self.jump_address
        relitive_jump = distance - 5  # size of this line
        packed_relitive_jump = struct.pack("<i", relitive_jump)
        return b"\xE9" + packed_relitive_jump

    def get_hook_bytecode(self) -> bytes:
        self.set_mouse_pos_addr()
        packed_mouse_pos_addr = pack_to_int_or_longlong(self.mouse_pos_addr)

        # fmt: off
        bytecode = (
            b"\x50"  # push rax
            b"\x48\xA1" + packed_mouse_pos_addr +  # mov rax, mouse_pos
            b"\x48\x89\x01"  # mov [rcx], rax
            b"\x58"  # pop rax
            b"\xC3"  # ret
        )
        # fmt: on

        return bytecode

    def get_pattern(self) -> Tuple[re.Pattern, str]:
        return (
            re.compile(rb"[\xBA\xE9]....\x44\x8D\x42\x7E\x48\xFF\x25\xF8\xB0\x06\x00"),
            "user32.dll",
        )

    def unhook(self):
        super().unhook()
        self.free_mouse_pos_addr()
        for bool_addr in self.toggle_bool_addrs:
            self.memory_handler.process.write_uchar(bool_addr, 0)



import asyncio
import functools
import struct
from collections import defaultdict

import pymem

from wizwalker import HookAlreadyActivated, utils, HookNotActive
from .hooks import (
    PlayerHook,
    QuestHook,
    PlayerStatHook,
    BackpackStatHook,
    DuelHook,
    MouselessCursorMoveHook,
)


def uses_hook(hook: str):
    def decorator(function):
        @functools.wraps(function)
        async def wrapped(*args, **kwargs):
            memory_handler = args[0]
            if not memory_handler.active_hooks[hook]:
                raise HookNotActive(hook)

            return await function(*args, **kwargs)

        return wrapped

    return decorator


def register_hook(hook: str):
    def decorator(function):
        @functools.wraps(function)
        async def wrapped(*args, **kwargs):
            memory_handler = args[0]
            if memory_handler.active_hooks[hook]:
                raise HookAlreadyActivated(hook)

            memory_handler.set_hook_active(hook)

            return await function(*args, **kwargs)

        return wrapped

    return decorator


class MemoryHandler:
    """
    Handles the internal memory calls, not for public use
    """

    def __init__(self, pid: int):
        self.process = pymem.Pymem()
        self.process.open_process_from_id(pid)
        self.process.check_wow64()

        self.player_struct_addr = None
        self.quest_struct_addr = None
        self.player_stat_addr = None
        self.backpack_stat_addr = None
        self.move_lock_addr = None
        self.current_duel_addr = None

        self.mouse_pos = None

        self.hooks = []
        self.active_hooks = defaultdict(lambda: False)

    @utils.executor_function
    def close(self):
        """
        Closes MemoryHandler, closing all hooks
        """
        for hook in self.hooks:
            try:
                hook.unhook()
            except pymem.exception.MemoryWriteError:
                # TODO: error here
                pass

        self.active_hooks = defaultdict(lambda: False)

    def set_hook_active(self, hook):
        self.active_hooks[hook] = True

    def hook_functs(self):
        hooks = {}
        # Couldn't get anything else working for this, other than manually setting
        # some attr on each hook method
        for thing in dir(self):
            if thing.startswith("hook_") and not any(
                thing.endswith(i) for i in ("all", "functs")
            ):
                hooks[thing.replace("hook_", "")] = getattr(self, thing)

        return hooks

    def is_hook_active(self, hook):
        return self.active_hooks[hook]

    @uses_hook("duel")
    @utils.executor_function
    def is_in_battle(self):
        """True if in battle; False if None"""
        current_battle_addr = self.process.read_longlong(self.current_duel_addr)
        try:
            duel_phase = self.process.read_int(current_battle_addr + 160)
        except pymem.exception.MemoryReadError:
            return None
        else:
            return duel_phase in (0, 1, 2, 3, 4, 5, 6)

    @uses_hook("mouseless_cursor_move")
    @utils.executor_function
    def write_mouse_position(self, x: int, y: int):
        self.process.write_int(self.mouse_pos, x)
        self.process.write_int(self.mouse_pos + 4, y)

    @uses_hook("player_struct")
    @utils.executor_function
    def read_player_base(self):
        return self.process.read_longlong(self.player_struct_addr)

    @uses_hook("player_stat_struct")
    @utils.executor_function
    def read_player_stat_base(self):
        return self.process.read_longlong(self.player_stat_addr)

    @uses_hook("backpack_stat_struct")
    @utils.executor_function
    def read_backpack_stat_base(self):
        return self.process.read_int(self.backpack_stat_addr)

    @uses_hook("quest_struct")
    @utils.executor_function
    def read_quest_base(self):
        return self.process.read_longlong(self.quest_struct_addr)

    @uses_hook("player_struct")
    @utils.executor_function
    def read_xyz(self):
        player_struct = self.process.read_longlong(self.player_struct_addr)
        try:
            x = self.process.read_float(player_struct + 0x58)  # was 2c
            y = self.process.read_float(player_struct + 0x58 + 4)
            z = self.process.read_float(player_struct + 0x58 + 4 + 4)
        except pymem.exception.MemoryReadError:
            return None
        else:
            return utils.XYZ(x, y, z)

    @uses_hook("player_struct")
    @utils.executor_function
    def set_xyz(self, x=None, y=None, z=None):
        player_struct = self.process.read_longlong(self.player_struct_addr)
        try:
            if x is not None:
                self.process.write_float(player_struct + 0x58, x)
            if y is not None:
                self.process.write_float(player_struct + 0x58 + 4, y)
            if z is not None:
                self.process.write_float(player_struct + 0x58 + 4 + 4, z)
        except pymem.exception.MemoryWriteError:
            return False
        else:
            return True

    @uses_hook("player_struct")
    @utils.executor_function
    def read_player_yaw(self):
        player_struct = self.process.read_longlong(self.player_struct_addr)
        try:
            return self.process.read_float(player_struct + 0x58 + 20)
        except pymem.exception.MemoryReadError:
            return None

    @uses_hook("player_struct")
    @utils.executor_function
    def set_player_yaw(self, yaw):
        player_struct = self.process.read_longlong(self.player_struct_addr)
        try:
            self.process.write_float(player_struct + 0x58 + 20, yaw)
        except pymem.exception.MemoryWriteError:
            return False
        else:
            return True

    @uses_hook("player_struct")
    @utils.executor_function
    def read_player_pitch(self):
        player_struct = self.process.read_longlong(self.player_struct_addr)
        try:
            return self.process.read_float(player_struct + 0x58 + 12)
        except pymem.exception.MemoryReadError:
            return None

    @uses_hook("player_struct")
    @utils.executor_function
    def set_player_pitch(self, pitch):
        player_struct = self.process.read_longlong(self.player_struct_addr)
        try:
            self.process.write_float(player_struct + 0x58 + 12, pitch)
        except pymem.exception.MemoryWriteError:
            return False
        else:
            return True

    @uses_hook("player_struct")
    @utils.executor_function
    def read_player_roll(self):
        player_struct = self.process.read_longlong(self.player_struct_addr)
        try:
            return self.process.read_float(player_struct + 0x58 + 16)
        except pymem.exception.MemoryReadError:
            return None

    @uses_hook("player_struct")
    @utils.executor_function
    def set_player_roll(self, roll):
        player_struct = self.process.read_longlong(self.player_struct_addr)
        try:
            self.process.write_float(player_struct + 0x58 + 16, roll)
        except pymem.exception.MemoryWriteError:
            return False
        else:
            return True

    @uses_hook("player_struct")
    @utils.executor_function
    def read_player_scale(self):
        player_struct = self.process.read_longlong(self.player_struct_addr)
        try:
            return self.process.read_float(player_struct + 0x58 + 24)
        except pymem.exception.MemoryReadError:
            return None

    @uses_hook("player_struct")
    @utils.executor_function
    def set_player_scale(self, scale):
        player_struct = self.process.read_longlong(self.player_struct_addr)
        try:
            self.process.write_float(player_struct + 0x58 + 24, scale)
        except pymem.exception.MemoryWriteError:
            return False
        else:
            return True

    @uses_hook("quest_struct")
    @utils.executor_function
    def read_quest_xyz(self):
        quest_struct = self.process.read_longlong(self.quest_struct_addr)
        try:
            x = self.process.read_float(quest_struct)
            y = self.process.read_float(quest_struct + 0x4)
            z = self.process.read_float(quest_struct + 0x8)

            return utils.XYZ(x, y, z)

        except pymem.exception.MemoryReadError:
            return None

    @uses_hook("player_stat_struct")
    @utils.executor_function
    def read_player_level(self):
        stat_addr = self.process.read_longlong(self.player_stat_addr)
        try:
            return self.process.read_int(stat_addr + 308)
        except pymem.exception.MemoryReadError:
            return None

    @uses_hook("player_stat_struct")
    @utils.executor_function
    def read_player_gardening_level(self):
        stat_addr = self.process.read_longlong(self.player_stat_addr)
        try:
            return self.process.read_uchar(stat_addr + 824)
        except pymem.exception.MemoryReadError:
            return None

    @uses_hook("player_stat_struct")
    @utils.executor_function
    def read_player_gardening_experience(self):
        stat_addr = self.process.read_longlong(self.player_stat_addr)
        try:
            return self.process.read_int(stat_addr + 828)
        except pymem.exception.MemoryReadError:
            return None

    @uses_hook("player_stat_struct")
    @utils.executor_function
    def read_player_fishing_level(self):
        stat_addr = self.process.read_longlong(self.player_stat_addr)
        try:
            return self.process.read_uchar(stat_addr + 853)
        except pymem.exception.MemoryReadError:
            return None

    @uses_hook("player_stat_struct")
    @utils.executor_function
    def read_player_fishing_experience(self):
        stat_addr = self.process.read_longlong(self.player_stat_addr)
        try:
            return self.process.read_int(stat_addr + 856)
        except pymem.exception.MemoryReadError:
            return None

    @uses_hook("player_stat_struct")
    @utils.executor_function
    def read_player_health(self):
        stat_addr = self.process.read_longlong(self.player_stat_addr)
        try:
            return self.process.read_int(stat_addr + 0x68)
        except pymem.exception.MemoryReadError:
            return None

    @uses_hook("player_stat_struct")
    @utils.executor_function
    def read_player_health_max(self):
        stat_addr = self.process.read_longlong(self.player_stat_addr)
        try:
            base = self.process.read_int(stat_addr + 80)
            bonus = self.process.read_int(stat_addr + 208)
            return base + bonus
        except pymem.exception.MemoryReadError:
            return None

    @uses_hook("player_stat_struct")
    @utils.executor_function
    def read_player_mana(self):
        stat_addr = self.process.read_longlong(self.player_stat_addr)
        try:
            return self.process.read_int(stat_addr + 0x78)
        except pymem.exception.MemoryReadError:
            return None

    @uses_hook("player_stat_struct")
    @utils.executor_function
    def read_player_mana_max(self):
        stat_addr = self.process.read_longlong(self.player_stat_addr)
        try:
            base = self.process.read_int(stat_addr + 84)
            bonus = self.process.read_int(stat_addr + 212)
            return base + bonus
        except pymem.exception.MemoryReadError:
            return None

    @uses_hook("player_stat_struct")
    @utils.executor_function
    def read_player_energy(self):
        stat_addr = self.process.read_longlong(self.player_stat_addr)
        try:
            return self.process.read_int(stat_addr + 0x64)
        except pymem.exception.MemoryReadError:
            return None

    @uses_hook("player_stat_struct")
    @utils.executor_function
    def read_player_potions(self):
        stat_addr = self.process.read_longlong(self.player_stat_addr)
        try:
            # this is a float for some reason
            # it's because you can partially fill potions with minigames -Forrest
            return self.process.read_float(stat_addr + 156)
        except pymem.exception.MemoryReadError:
            return None

    @uses_hook("player_stat_struct")
    @utils.executor_function
    def read_player_potions_max(self):
        stat_addr = self.process.read_longlong(self.player_stat_addr)
        try:
            return self.process.read_float(stat_addr + 152)
        except pymem.exception.MemoryReadError:
            return None

    @uses_hook("player_stat_struct")
    @utils.executor_function
    def read_player_gold(self):
        stat_addr = self.process.read_longlong(self.player_stat_addr)
        try:
            return self.process.read_int(stat_addr + 108)
        except pymem.exception.MemoryReadError:
            return None

    @uses_hook("backpack_stat_struct")
    @utils.executor_function
    def read_player_backpack_used(self):
        backpack_addr = self.process.read_longlong(self.backpack_stat_addr)
        try:
            return self.process.read_int(backpack_addr)
        except pymem.exception.MemoryReadError:
            return None

    @uses_hook("backpack_stat_struct")
    @utils.executor_function
    def read_player_backpack_total(self):
        backpack_addr = self.process.read_longlong(self.backpack_stat_addr)
        try:
            return self.process.read_int(backpack_addr + 0x4)
        except pymem.exception.MemoryReadError:
            return None

    @uses_hook("duel")
    @utils.executor_function
    def read_move_lock(self):
        current_battle_addr = self.process.read_longlong(self.current_duel_addr)
        try:
            duel_phase = self.process.read_int(current_battle_addr + 160)
        except pymem.exception.MemoryReadError:
            return None
        else:
            return duel_phase in (0, 1, 2, 3, 4, 5, 6)

    async def hook_all(self):
        hooks = []
        for hook in self.hook_functs().values():
            # Need the coroutine objects
            hooks.append(hook())

        # return_exceptions=True will make all exceptions return as results
        return await asyncio.gather(*hooks, return_exceptions=True)

    @register_hook("player_struct")
    @utils.executor_function
    def hook_player_struct(self):
        player_hook = PlayerHook(self)
        player_hook.hook()

        self.hooks.append(player_hook)
        self.player_struct_addr = player_hook.player_struct

    @register_hook("quest_struct")
    @utils.executor_function
    def hook_quest_struct(self):
        quest_hook = QuestHook(self)
        quest_hook.hook()

        self.hooks.append(quest_hook)
        self.quest_struct_addr = quest_hook.cord_struct

    @register_hook("player_stat_struct")
    @utils.executor_function
    def hook_player_stat_struct(self):
        player_stat_hook = PlayerStatHook(self)
        player_stat_hook.hook()

        self.hooks.append(player_stat_hook)
        self.player_stat_addr = player_stat_hook.stat_addr

    @register_hook("backpack_stat_struct")
    @utils.executor_function
    def hook_backpack_stat_struct(self):
        backpack_stat_hook = BackpackStatHook(self)
        backpack_stat_hook.hook()

        self.hooks.append(backpack_stat_hook)
        self.backpack_stat_addr = backpack_stat_hook.backpack_struct_addr

    # @register_hook("move_lock")
    # @utils.executor_function
    # def hook_move_lock(self):
    #     move_lock_hook = MoveLockHook(self)
    #     move_lock_hook.hook()
    #
    #     self.hooks.append(move_lock_hook)
    #     self.move_lock_addr = move_lock_hook.move_lock_addr

    @register_hook("duel")
    @utils.executor_function
    def hook_duel(self):
        duel_hook = DuelHook(self)
        duel_hook.hook()

        self.hooks.append(duel_hook)
        self.current_duel_addr = duel_hook.current_duel_addr

    @register_hook("mouseless_cursor_move")
    @utils.executor_function
    def hook_mouseless_cursor_move(self):
        mouseless_cursor_hook = MouselessCursorMoveHook(self)
        mouseless_cursor_hook.hook()

        self.hooks.append(mouseless_cursor_hook)
        self.mouse_pos = mouseless_cursor_hook.mouse_pos_addr

        self.process.write_longlong(self.mouse_pos, 0)



from typing import List, Union
from wizwalker import NotEnoughMana, NotEnoughPips, CardNotFound, CardAlreadyEnchanted


class Hand:
    def __init__(self, client: "wizwalker.Client"):
        self.client = client
        self._cards: List[Card] = []

    async def _update_card_data(self):
        pass
        # self._cards = new_data

    def _search_card(self, name):
        pass
        # raise CardNotFound(name)

    def _card_from_position(self, position):
        pass
        # raise CardNotFound(position)

    async def size(self) -> int:
        pass

    async def get_card(self, card_name_or_positon: Union[str, int]) -> "Card":
        """
        Get a card based on a name or position
        """
        if not isinstance(card_name_or_positon, (str, int)):
            raise ValueError(
                f"Card_name_or_position must be str or int not {card_name_or_positon.__class__!r}"
            )

        await self._update_card_data()

        if isinstance(card_name_or_positon, str):
            card = self._search_card(card_name_or_positon)

        else:
            card = self._card_from_position(card_name_or_positon)

        return Card()

    async def get_all_cards(self) -> list:
        await self._update_card_data()
        return self._cards


class Card:
    def __init__(self):
        self.position = None
        self.name = None
        self.pips = None
        self.shadow_pips = None
        self.is_enchanted = None

    async def enchant(self, card: "Card"):
        if card.is_enchanted:
            raise CardAlreadyEnchanted()

        # enchant code here
        card.is_enchanted = True
        raise NotImplementedError()

    async def target(self, combat_position: int):
        raise NotImplementedError()

    async def discard(self):
        raise NotImplementedError()



import struct
import zlib
from collections import namedtuple
from typing import List, Union
from pathlib import Path

import aiofiles

from .utils import executor_function, get_wiz_install

wad_file_info = namedtuple("wad_file_info", "offset, size, is_zip, crc, unzipped_size")


class WadFileInfo:
    def __init__(self, *, name, offset, size, is_zip, crc, unzipped_size):
        self.name = name
        self.offset = offset
        self.size = size
        self.is_zip = is_zip
        self.crc = crc
        self.unzipped_size = unzipped_size


class Wad:
    def __init__(self, path: Union[Path, str]):
        if isinstance(path, str):
            path = Path(path)

        self.name = path.with_suffix("").name

        self.file_path = path
        if not self.file_path.exists():
            raise ValueError(f"{self.file_path} not found.")

        self._file_list = []
        self._refreshed_once = False
        self._open = False
        self._file_pointer = None

    @classmethod
    def from_game_data(cls, name: str):
        """
        Get a Wad file from game installation dir

        Args:
            name: name of the wad
        """
        if not name.endswith(".wad"):
            name += ".wad"

        file_path = get_wiz_install() / "Data" / "GameData" / name
        return cls(file_path)

    def __repr__(self):
        return f"<Wad {self.name=}>"

    async def size(self) -> int:
        """
        Total size of this wad
        """
        if not self._open:
            await self.open()

        return sum(file.size for file in self._file_list)

    async def names(self) -> List[str]:
        """
        List of all file names in this wad
        """
        if not self._open:
            await self.open()

        return [file.name for file in self._file_list]

    async def open(self):
        # noinspection PyTypeChecker
        # TODO: why is this stored but not used in half the methods
        self._file_pointer = open(self.file_path, "rb")
        await self._refresh_journal()
        self._open = True

    def close(self):
        self._file_pointer.close()
        self._open = False

    @executor_function
    def _refresh_journal(self):
        if self._refreshed_once:
            return

        self._refreshed_once = True

        fp = self._file_pointer

        # KIWAD id string
        fp.seek(5)
        version = struct.unpack("<l", fp.read(4))[0]
        file_num = struct.unpack("<l", fp.read(4))[0]

        if version >= 2:
            fp.read(1)

        for _ in range(file_num):
            offset = struct.unpack("<l", fp.read(4))[0]
            size = struct.unpack("<l", fp.read(4))[0]
            zsize = struct.unpack("<l", fp.read(4))[0]
            is_zip = struct.unpack("?", fp.read(1))[0]
            crc = struct.unpack("<l", fp.read(4))[0]
            name_length = struct.unpack("<l", fp.read(4))[0]
            name = (fp.read(name_length)).decode("utf-8")[:-1]

            self._file_list.append(
                WadFileInfo(
                    name=name,
                    offset=offset,
                    size=size,
                    is_zip=is_zip,
                    crc=crc,
                    unzipped_size=zsize,
                )
            )

    async def get_file(self, name: str) -> bytes:
        """
        Get the data contents of the named file

        Args:
            name: name of the file to get
        """
        if not self._open:
            await self.open()

        target_file = None
        for file in self._file_list:
            if file.name == name:
                target_file = file

        if not target_file:
            raise ValueError(f"File {name} not found.")

        async with aiofiles.open(self.file_path, "rb") as fp:
            await fp.seek(target_file.offset)
            raw_data = await fp.read(target_file.size)

            if target_file.is_zip:
                try:
                    data = zlib.decompress(raw_data)
                except zlib.error:
                    data = raw_data

            else:
                data = raw_data

        return data

    async def get_file_info(self, name: str) -> wad_file_info:
        """
        Gets a WadFileInfo for a named file

        Args:
            name: name of the file to get info on
        """
        if not self._open:
            await self.open()

        target_file = None
        for file in self._file_list:
            if file.name == name:
                target_file = file

        if not target_file:
            raise ValueError(f"File {name} not found.")

        return target_file

    async def unarchive(self, path: Union[Path, str]):
        """
        Unarchive a wad file into a directory

        Args:
            path: path to the directory to unpack the wad
        """
        if isinstance(path, str):
            path = Path(path)

        if not path.exists():
            raise ValueError(f"{path} does not exist.")

        if not path.is_dir():
            raise ValueError(f"{path} is not a directory.")

        for file in self._file_list:
            dirs = file.name.split("/")
            # not a base level file
            if len(dirs) != 1:
                current = path
                for next_dir in dirs[:-1]:
                    current = current / next_dir
                    if not current.exists():
                        current.mkdir()

            file_path = path / file.name
            file_data = await self.get_file(file.name)

            async with aiofiles.open(file_path, "wb") as fp:
                await fp.write(file_data)

    @classmethod
    async def from_directory(self, path: Union[Path, str]):
        """
        Create a Wad object from a directory

        Args:
            path: Path to directory to archive
        """
        if isinstance(path, str):
            path = Path(path)

        if not path.exists():
            raise ValueError(f"{path} does not exist.")

        if not path.is_dir():
            raise ValueError(f"{path} is not a directory.")

        raise NotImplemented()



import asyncio
import ctypes.wintypes
import warnings
from functools import cached_property
from typing import Optional

from . import Keycode, utils
from .windows import MemoryHandler

from .constants import user32


WIZARD_SPEED = 580


class Client:
    """
    Represents a connected wizard client

    Args:
        window_handle: A handle to the window this client connects to
    """

    def __init__(self, window_handle: int):
        self.window_handle = window_handle
        self._memory = MemoryHandler(self.process_id)

        self.click_lock = None

    def __repr__(self):
        return f"<Client {self.window_handle=} {self.process_id=}>"

    async def close(self):
        """
        Closes this client; unhooking all active hooks
        """
        await self._memory.close()

    @cached_property
    def process_id(self) -> int:
        """
        This client's PID

        Returns:
            The pid
        """
        # https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-getwindowthreadprocessid
        pid = ctypes.wintypes.DWORD()
        user32.GetWindowThreadProcessId(self.window_handle, ctypes.byref(pid))
        return pid.value

    async def send_key(self, key: Keycode, seconds: float):
        await utils.timed_send_key(self.window_handle, key, seconds)

    async def click(
        self,
        x: int,
        y: int,
        *,
        right_click: bool = False,
        sleep_duration: float = 0.0,
        use_post: bool = False,
    ):
        """
        Send a click to a certain x and y
        x and y positions are relative to the top left corner of the screen

        Args:
            x: x to click at
            y: y to click at
            right_click: If the click should be a right click
            sleep_duration: How long to sleep between messages
            use_post: If PostMessage should be used instead of SendMessage
        """
        # prevent multiple clicks from happening at the same time
        if right_click:
            button_down_message = 0x204
        else:
            button_down_message = 0x201

        if use_post:
            send_method = user32.PostMessageW
        else:
            send_method = user32.SendMessageW

        if self.click_lock is None:
            self.click_lock = asyncio.Lock()

        async with self.click_lock:
            # TODO: test passing use_post
            await self.set_mouse_position(x, y)
            # mouse button down
            send_method(self.window_handle, button_down_message, 1, 0)
            if sleep_duration > 0:
                await asyncio.sleep(sleep_duration)
            # mouse button up
            send_method(self.window_handle, button_down_message + 1, 0, 0)

    def login(self, username: str, password: str):
        """
        Login to a client that is at the login screen

        Args:
            username: The username to login with
            password: The password to login with
        """
        utils.instance_login(self.window_handle, username, password)

    def get_hooks(self) -> list:
        """
        return a list of hook_names from the underlying MemoryHandler

        Returns:
            A list of avalible hooks
        """
        return self._memory.hook_functs().keys()

    async def activate_hooks(self, *hook_names: Optional[str]):
        """
        Activate a number of hooks or pass None/no args to activate all

        Args:
            hook_names: The hooks to activate

        Examples:
            .. code-block:: py

                # activates player_struct and player_stat_struct
                activate_hooks("player_struct", "player_stat_struct")

                # activates all hooks
                activate_hooks()

        """
        if not hook_names:
            await self._memory.hook_all()

        else:
            for hook_name in hook_names:
                hook = getattr(self._memory, "hook_" + hook_name, None)
                if not hook:
                    raise ValueError(f"{hook_name} is not a valid hook")

                await hook()

    async def goto(
        self,
        x: float,
        y: float,
        *,
        speed_multiplier: float = 1.0,
        use_nodes: bool = False,
    ):
        """
        Moves the player to a specific x and y

        Args:
            x: X to move to
            y: Y to move to
            speed_multiplier: Multiplier for speed (for mounts) i.e. 1.4 for 40%
            use_nodes: If node date should be used, currently WIP
        """
        if use_nodes is False:
            await self._to_point(x, y, speed_multiplier)
        else:
            raise NotImplemented("Node data not implemented")

    async def _to_point(self, x, y, speed_multiplier=1.0):
        current_xyz = await self.xyz()
        target_xyz = utils.XYZ(x, y, current_xyz.z)
        distance = current_xyz - target_xyz
        move_seconds = distance / (WIZARD_SPEED * speed_multiplier)
        yaw = utils.calculate_perfect_yaw(current_xyz, target_xyz)

        await self.set_yaw(yaw)
        await utils.timed_send_key(self.window_handle, Keycode.W, move_seconds)

    async def set_mouse_position(
        self,
        x: int,
        y: int,
        *,
        convert_from_client: bool = True,
        use_post: bool = False,
    ):
        """
        Set's the mouse position to a certain x y relative to the
        top left corner of the client

        Args:
            x: x to set
            y: y to set
            convert_from_client: If the position should be converted from client to screen
            use_post: If PostMessage should be used instead of SendMessage
        """
        if use_post:
            send_method = user32.PostMessageW
        else:
            send_method = user32.SendMessageW

        if convert_from_client:
            point = ctypes.wintypes.tagPOINT(x, y)

            # https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-clienttoscreen
            if user32.ClientToScreen(self.window_handle, ctypes.byref(point)) == 0:
                raise RuntimeError("Client to screen conversion failed")

            # same point structure is overwritten by ClientToScreen; these are also ints and not
            # c_longs for some reason?
            x = point.x
            y = point.y

        res = await self._memory.write_mouse_position(x, y)
        # position doesn't matter here; sending mouse move
        # mouse move is here so that items are highlighted
        send_method(self.window_handle, 0x200, 0, 0)
        return res

    async def teleport(
        self, x: float = None, y: float = None, z: float = None, yaw: float = None
    ) -> bool:
        """
        Teleport the player to a set x, y, z

        Args:
            x: X to teleport to or None to not change
            y: Y to teleport to or None to not change
            z: Z to teleport to or None to not change
            yaw: yaw to set or None to not change

        Raises:
            RuntimeError: player_struct hook not active

        Returns:
            True if telelporting succseeded, False otherwise
        """
        res = await self.set_xyz(x=x, y=y, z=z)

        if yaw is not None:
            res = res and await self.set_yaw(yaw)

        return res

    async def xyz(self) -> Optional[utils.XYZ]:
        """
        Player xyz

        Raises:
            RuntimeError: player_struct hook not active

        Returns:
            XYZ namedtuple or None if hooked function hasn't run yet
        """
        return await self._memory.read_xyz()

    async def set_xyz(self, x: float, y: float, z: float):
        """
        Set the player xyz

        Raises:
            RuntimeError: player_struct hook not active

        Returns:
            XYZ namedtuple or None if hooked function hasn't run yet
        """
        return await self._memory.set_xyz(x=x, y=y, z=z)

    async def yaw(self) -> Optional[float]:
        """
        Player yaw

        Raises:
            RuntimeError: player_struct hook not active

        Returns:
            yaw float or None if hooked function hasn't run yet
        """
        return await self._memory.read_player_yaw()

    async def set_yaw(self, yaw: float) -> bool:
        """
        Set the player yaw

        Raises:
            RuntimeError: player_struct hook not active

        Return:
            True if value was set, False otherwise
        """
        return await self._memory.set_player_yaw(yaw)

    async def roll(self) -> Optional[float]:
        """
        Player roll

        Raises:
            RuntimeError: player_struct hook not active

        Returns:
            roll float or None if hooked function hasn't run yet
        """
        return await self._memory.read_player_roll()

    async def set_roll(self, roll: float) -> bool:
        """
        Set the player roll

        Raises:
            RuntimeError: player_struct hook not active

        Return:
            True if value was set, False otherwise
        """
        return await self._memory.set_player_roll(roll)

    async def pitch(self) -> Optional[float]:
        """
        Player pitch

        Raises:
            RuntimeError: player_struct hook not active

        Returns:
            pitch float or None if hooked function hasn't run yet
        """
        return await self._memory.read_player_pitch()

    async def set_pitch(self, pitch: float) -> bool:
        """
        Set the player pitch

        Raises:
            RuntimeError: player_struct hook not active

        Return:
            True if value was set, False otherwise
        """
        return await self._memory.set_player_roll(pitch)

    async def scale(self) -> Optional[float]:
        """
        Player scale

        Raises:
            RuntimeError: player_struct hook not active

        Returns:
            scale float or None if hooked function hasn't run yet
        """
        return await self._memory.read_player_scale()

    async def set_scale(self, scale: float) -> bool:
        """
        Set the player scale

        Raises:
            RuntimeError: player_struct hook not active

        Return:
            True if value was set, False otherwise
        """
        return await self._memory.set_player_scale(scale)

    async def quest_xyz(self) -> Optional[utils.XYZ]:
        """
        Quest xyz

        Raises:
            RuntimeError: quest_struct hook not active

        Return:
            XYZ namedtuple or None if hooked function hasn't run yet
        """
        return await self._memory.read_quest_xyz()

    async def move_lock(self) -> Optional[bool]:
        """
        Player move lock; weither or not the player is locked
        in combat/dialog

        Raises:
            RuntimeError: move_lock hook not active

        Return:
            move look bool or None if the function has not run yet
        """
        return await self._memory.read_move_lock()

    async def level(self) -> Optional[int]:
        """
        Player level

        Raises:
            RuntimeError: player_stat_struct hook not active

        Return:
            level int or None if hooked function hasn't run yet
        """
        return await self._memory.read_player_level()

    async def gardening_level(self) -> Optional[int]:
        """
        Player gardening level

        Raises:
            RuntimeError: player_stat_struct hook not active

        Return:
            gardening level int or None if hooked function hasn't run yet
        """
        return await self._memory.read_player_gardening_level()

    async def gardening_experience(self) -> Optional[int]:
        """
        Player gardening experience

        Raises:
            RuntimeError: player_stat_struct hook not active

        Return:
            gardening experience int or None if hooked function hasn't run yet
        """
        return await self._memory.read_player_gardening_experience()

    async def fishing_level(self) -> Optional[int]:
        """
        Player fishing level

        Raises:
            RuntimeError: player_stat_struct hook not active

        Return:
            fishing level int or None if hooked function hasn't run yet
        """
        return await self._memory.read_player_fishing_level()

    async def fishing_experience(self) -> Optional[int]:
        """
        Player fishing_experience

        Raises:
            RuntimeError: player_stat_struct hook not active

        Return:
            fishing experience int or None if hooked function hasn't run yet
        """
        return await self._memory.read_player_fishing_experience()

    async def health(self) -> Optional[int]:
        """
        Player health

        Raises:
            RuntimeError: player_stat_struct hook not active

        Return:
            health int or None if hooked function hasn't run yet
        """
        return await self._memory.read_player_health()

    async def max_health(self) -> Optional[int]:
        """
        Player max health

        Raises:
            RuntimeError: player_stat_struct hook not active

        Return:
            max health int or None if hooked function hasn't run yet
        """
        return await self._memory.read_player_health_max()

    async def mana(self) -> Optional[int]:
        """
        Player mana

        Raises:
            RuntimeError: player_stat_struct hook not active

        Return:
            mana int or None if hooked function hasn't run yet
        """
        return await self._memory.read_player_mana()

    async def max_mana(self) -> Optional[int]:
        """
        Player max mana

        Raises:
            RuntimeError: player_stat_struct hook not active

        Return:
            max mana int or None if hooked function hasn't run yet
        """
        return await self._memory.read_player_mana_max()

    async def energy(self) -> Optional[int]:
        """
        Player max energy

        Raises:
            RuntimeError: player_stat_struct hook not active

        Return:
            max energy int or None if hooked function hasn't run yet
        """
        return await self._memory.read_player_energy()

    async def potions(self) -> Optional[float]:
        """
        Player full potions

        Raises:
            RuntimeError: player_stat_struct hook not active

        Return:
            potions float or None if hooked function hasn't run yet
        """
        return await self._memory.read_player_potions()

    async def max_potions(self) -> Optional[float]:
        """
                Player full potions

        Raises:
            RuntimeError: player_stat_struct hook not active

        Return:
            max potions float or None if hooked function hasn't run yet
        """
        return await self._memory.read_player_potions_max()

    # TODO: remove in 1.0
    async def potions_alt(self) -> Optional[float]:
        """
        copatibility function for old scripts; removed in 1.0
        """
        warnings.warn(
            "potions_alt is depreciated and will be removed in 1.0; please use potions instead",
            DeprecationWarning,
        )
        return await self._memory.read_player_potions()

    async def gold(self) -> Optional[int]:
        """
        Player gold

        Raises:
            RuntimeError: player_stat_struct hook not active

        Return:
            gold int or None if hooked function hasn't run yet
        """
        return await self._memory.read_player_gold()

    async def backpack_space_used(self) -> Optional[int]:
        """
        Player backpack used space

        Raises:
            RuntimeError: backpack_struct hook not active

        Return:
            used backpack space int or None if hooked function hasn't run yet
        """
        return await self._memory.read_player_backpack_used()

    async def backpack_space_total(self) -> Optional[int]:
        """
        Player backpack total space

        Raises:
            RuntimeError: backpack_struct hook not active

        Return:
            backpack total space int or None if hooked function hasn't run yet
        """
        return await self._memory.read_player_backpack_total()
